/* Computes direct Carleman Bilinearization using sparsity oriented representations
 
 Compile with mex cbdsp.c  
 
 
CBVK Toolbox
JHMG, PMSB 2016

*/

#include <string.h>
#include "mex.h"


#define NMAX 200		/* Maximum number of non-null columns of bilinear model */

struct tnode{			/* Tree node */
  double itylr;			/* Taylor expansion term index */
  int iout;			    /* Ouput vector index */
  struct tnode *left;	/* Left child */
  struct tnode *right;  /* Right child node */
};

struct tree{
    struct tnode *root;
    int nptr;
    struct tnode nodes[NMAX];
};    

/* Global variables */

int num_adds;

 struct tree *tsf, *tsg;
/*struct tree tsf[14641]; 
struct tree tsg[14641]; */

/*%%%%%%%%%%% AddTree %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

/* Function addtree: adds node with index itl and coefficient cfc at p or below p.
 * Binary tree similar to K&R p.139, but without dynamical alocation */

struct tnode *addtree(struct tree *t, struct tnode *p, double itl, double cfc, double *outi, double *outc){
  int cond;

  if (p == NULL){                   /* A new index was received */
    p = &(t->nodes[t->nptr]);		/* Pointer to new node */
    p->itylr = itl;
    p->iout = t->nptr;
    outi[t->nptr] = itl;            /* Initialize output vectors */
    outc[t->nptr] = cfc;		
    p->left = p->right = NULL;

    t->nptr = t->nptr + 1;          /* Prepare for next new index */

    if (t->nptr >= NMAX){
      mexErrMsgIdAndTxt( "cbdsp:nmax", "Maximum number of non-null columns exceeded");
    }

  }
  else if (cond = ((p->itylr - itl) == 0)){ /* Index of existing expansion */
    outc[p->iout] += cfc;                   /* Adds coeffcient to existing one */
    num_adds++;
  }
  else if (cond <0)		/* Searches for smaller indices */
    p->left = addtree(t, p->left, itl, cfc, outi, outc);
  else				/* or larger */
    p->right = addtree(t, p->right, itl, cfc, outi, outc);
  return p;    
}

/*%%%%%%%%%%% DisplayMatrix %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

void DisplayMatrix(char *Name, double *Data, int M, int N)
{ /* Display matrix data */
int m, n;
mexPrintf("%s = \n", Name);
for(m = 0; m < M; m++, mexPrintf("\n"))
for(n = 0; n < N; n++)
mexPrintf("%8.4f ", Data[m + M*n]);
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*%%%%%%%%%%% Main function    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
   
  double *gi, *gc, *fi, *fc, *Pp, *mp, *MPp, *allterms, *numtermsF, *numtermsG, *NAp;
  int P, m, MP;

  /* Get inputs */
  fi = mxGetPr(prhs[0]);	
  fc = mxGetPr(prhs[1]);	
  gi = mxGetPr(prhs[2]);	
  gc = mxGetPr(prhs[3]);
  mp = mxGetPr(prhs[4]);
  m = *mp;  
  Pp = mxGetPr(prhs[5]);
  P = *Pp;
  MPp = mxGetPr(prhs[6]);
  MP = *MPp;
  allterms = mxGetPr(prhs[7]);
  numtermsF = mxGetPr(prhs[8]);
  numtermsG = mxGetPr(prhs[9]);
  
  double *Fi, *Fc, *Gi, *Gc;
  
  /* Create outputs */
  plhs[0] = mxCreateDoubleMatrix(NMAX,MP, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(NMAX,MP, mxREAL);
  Fi = mxGetPr(plhs[0]);	
  Fc = mxGetPr(plhs[1]);	
  plhs[2] = mxCreateDoubleMatrix(NMAX,MP, mxREAL);
  plhs[3] = mxCreateDoubleMatrix(NMAX,MP, mxREAL);

  plhs[4] = mxCreateDoubleMatrix(1,1, mxREAL);
  NAp = mxGetPr(plhs[4]);

  Gi = mxGetPr(plhs[2]);	
  Gc = mxGetPr(plhs[3]);	
  
  int n, i, mmax, nt1, nt2;
  
  /* Copy expansion lines */
  mmax =  mxGetM(prhs[0]);	
  for (n=0;n<m;n++){
    for(i=0;i<mmax;i++){
      Fi[nt1=n*NMAX+i]=fi[nt2=n*mmax+i];
      Fc[nt1]=fc[nt2];

    }
  }  
  mmax =  mxGetM(prhs[2]);	
  for (n=0;n<m;n++){
    for(i=0;i<mmax;i++){
      Gi[nt1=n*NMAX+i]=gi[nt2=n*mmax+i];
      Gc[nt1]=gc[nt2];

    }
  }  
  /* Auxiliary variables */
  int k, l0, p, termsp, ct, c0i, c0f, totalcols, c0bf, c0bg, pot_esq, pot_dir;
  int cb, cr, ce, itl, cl, itl_term, pos_prox_f, pos_prox_g, lin;
  double cf_term;
  
  /* Alocation of memory for structure */
  int termsP;
  termsP = allterms[P-1];
  tsf = (struct tree *) calloc(termsP, sizeof(struct tree));
  tsg = (struct tree *) calloc(termsP, sizeof(struct tree));
  /*
  for (ct=0;ct<termsP;ct++) {
      tsf[ct] = (struct tree *) malloc(sizeof(struct tree));
      tsf[ct] = (struct tree *) malloc(sizeof(struct tree));
      for (k=0;k<NMAX;k++) {
          tsf[ct].nodes[k] = malloc(sizeof struct tnode);
          tsg[ct].nodes[k] = malloc(sizeof struct tnode);
      }
  }  */
  
  int ini_lin, next_el, desloc_term, ini_bloco_l, ini_bloco_col;
  
  num_adds = 0;
  
  /* Main loop */
  l0 = allterms[0];
  for (p=2; p<=P; p++) {
      termsp = allterms[p-1];
      /* Initialization of p order trees */
      for (ct=0;ct<termsp;ct++) {
          tsf[ct].nptr = 0;
          tsf[ct].root = NULL;
          tsg[ct].nptr = 0;
          tsg[ct].root = NULL;
      }
      /* Computation of m^0, ..., m^(p-1) */
      int mpowers[p];
      mpowers[0] = 1;
      for (i=1; i<p; i++) {
          mpowers[i] = mpowers[i-1]*m;
      }   
      /* Row loop */
      ini_lin = 0;
      for (cl=0; cl<m; cl++) {
          pos_prox_f = 0;    /* Points no next non-null element of the row of F */
          pos_prox_g = 0;    /* Points no next non-null element of the row of G */
          /* Block loop for each order */
          for (i=1; i<=P-p+1; i++) {
              c0i = numtermsF[i-1];     /* Initial element of the i-th input block */
              c0f = numtermsF[i];       /* Final element of the i-th input block */
              c0bf = numtermsF[i+p-2];  /* Initial element of the i-th outpur block in F */
              totalcols = c0f-c0i;
              /* Computation of F: looking for next element in the row */
              while ((pos_prox_f < NMAX) && (Fi[next_el = ini_lin+pos_prox_f] <= c0f) && (Fi[next_el] > 0)) {
                  itl_term = Fi[next_el];  /* Taylor index of the next element */
                  cf_term = Fc[next_el];   /* Coefficient of the next element  */
                  /* Terms loop */
                  for (k=0; k<p; k++) {   
                      pot_esq = mpowers[k];
                      pot_dir = mpowers[p-1-k];       
                      desloc_term = (itl_term-c0i)*pot_dir;
                      /* Sub-block loop*/
                      for (cb=0; cb<pot_esq; cb++){ 
                          /* Repetion loop */
                          ini_bloco_l = (cb*m+cl)*pot_dir; 
                          ini_bloco_col = cb*pot_dir*totalcols + desloc_term;
                          for (cr=0; cr<pot_dir; cr++){
                              lin = ini_bloco_l+cr;
                              itl = ini_bloco_col+cr;
                              tsf[lin].root = addtree(&tsf[lin],tsf[lin].root,c0bf+itl,
                                               cf_term, 
                                               &Fi[(l0+lin)*NMAX], 
                                               &Fc[(l0+lin)*NMAX]);    
                          }
                      }
                  }                      
                  pos_prox_f++;
              }           
              c0i = numtermsG[i-1];     /* Initial element of the i-th input block*/
              c0f = numtermsG[i];       /* Final element of the i-th input block*/
              c0bg = numtermsG[i+p-2];  /* Initial element of the i-th output block in G */              
              totalcols = c0f-c0i;
              /* Computation of G: looking for next element in the row */
              while ((pos_prox_g < NMAX) && (Gi[next_el = ini_lin+pos_prox_g] <= c0f) && (Gi[next_el] > 0)) {
                  itl_term = Gi[next_el];
                  cf_term = Gc[next_el];
                  /*mexPrintf("\n p=%1d, i=%1d, Linha=%1d, itl_term=%3d, cf_term=%8.4f \n",p,i,cl,itl_term,cf_term);*/
                  /* Term loop */
                  for (k=0;k<p;k++) {   
                      pot_esq = mpowers[k];
                      pot_dir = mpowers[p-1-k];
                      desloc_term = (itl_term-c0i)*pot_dir;
                      /* Sub-block loop */
                      for (cb=0; cb<pot_esq; cb++){ 
                          /* Repetition loop */
                          ini_bloco_l = (cb*m+cl)*pot_dir; 
                          ini_bloco_col = cb*pot_dir*totalcols + desloc_term;                          
                          for (cr=0; cr<pot_dir; cr++){
                              lin = ini_bloco_l+cr;
                              itl = ini_bloco_col+cr;
                              tsg[lin].root = addtree(&tsg[lin],tsg[lin].root,c0bg+itl,
                                               cf_term, 
                                               &Gi[(l0+lin)*NMAX], 
                                               &Gc[(l0+lin)*NMAX]);
                              /*if (p < 3) {
                                 mexPrintf("> Saida: Linha=%4d, Coluna=%4d, c0bg=%4d \n",l0+lin,c0bg+itl,c0bg);
                              } */                             
                          }
                      }
                  }                      
                  pos_prox_g++;
              }              
          }
          ini_lin += NMAX;
      }
      l0 = l0+termsp;
  }
  /* mexPrintf("> Total de somas=%10d \n",num_adds); */
  *NAp = (double) num_adds;
  
  free(tsf);
  free(tsg);


}    
