/* Calculates non-redundant Carleman Bilinearization using sparsity
 
 Compile with mex cbnrsp.c
  
CBVK Toolbox
JHMG, PMSB 2016

*/
#include "mex.h"


#define NMAX 50			/* Maximum number of bilinearization non-null terms */


struct tnode{			/* Tree node */
  double itylr;			/* Taylor expansion index */
  int iout;             /* Output vector index    */
  struct tnode *left;	/* Left child node */
  struct tnode *right;	/* Right child node */
};



/* Global variables */

struct tnode nodes[NMAX];	/* Node array */
int nptr = 0;               /* Node pointer counter  */
size_t Np,Np1,ND;           /* Dimension of matrices */
int num_adds = 0;           /* >>> Problem with this */




/* Function addtree: adds node with index itl and coefficient cfc at p or 
 * under p. Binary tree is similar to K&R p. 139 but not using dynamical 
 * alocation of node space. 
 * For greater simplicity, uses global variables nodes and ntpr, when 
 * being called. As a consequence, its use cannot be interleaved between
 * more than one vector combination */

struct tnode *addtree(struct tnode *p, double itl, double cfc, double *outi, double *outc){
  int cond;

  if (p == NULL){		/* A new index was received */
    p = &nodes[nptr];	/* Assigns address of new node to pointer */
    p->itylr = itl;
    p->iout = nptr;
    outi[nptr] = itl;	/* Initializes output vectors */
    outc[nptr] = cfc;		
    p->left = p->right = NULL;

    nptr = nptr + 1;	/* Leaves all prepared for next new index */

    if (nptr >= NMAX){
       mexErrMsgIdAndTxt( "MATLAB:carlbilin:NMAX", "Maximum number of expansion terms exceeded");
    }

  }
  else if (cond = ((p->itylr - itl) == 0)){ /* Existing expansion index */
    outc[p->iout] += cfc;		    /* Adds coefficient to the existing one*/
    /* num_adds++; */
  }
  else if (cond <0)                 /* Searches for smaller indices  */
    p->left = addtree(p->left, itl, cfc, outi, outc);
  else                              /* or larger */
    p->right = addtree(p->right, itl, cfc, outi, outc);
  return p;
    
}



/* Function called from carlbilinsp, having nrhs and nlhs input and output 
 * arguments (matrices) para entrada e sada, respectively. 
 * Pointers prhs[nrhs] e plhs[nlhs] are for structures associated to these
 * matrices. These structures contain a pointer to the elements of the 
 * corresponding matrix, organized column-wise. Therefore, judging
 * from the Matlab refbook examples, the structure of a matrix is not the 
 * same in Matlab as in C, where matrices are accessed by a vector containing
 * a pointer to each row. */


void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{


  double *fi,*fc,*gi,*gc,*Fi,*Fc,*Gi,*Gc,*K,*D,*Dg,*bp,*mp,*Npp,ft,ft1,*NDp,*NAp;
  struct tnode *root;
  int i,n10,n20,n0,n,m,mmax,k1,k2,nt1,nt2;
  

  /* Extraction of input argument pointers */

  fi = mxGetPr(prhs[0]);	/* Indices of the non-null coefficients of the Taylor expansion of f */

  fc = mxGetPr(prhs[1]);	/* Non-null coefficients of the Taylor expansion of f */
  gi = mxGetPr(prhs[2]);	/* Same, g */
  gc = mxGetPr(prhs[3]);	
  mp = mxGetPr(prhs[4]);	/* Dimension m of state vector */
  Npp = mxGetPr(prhs[5]);	/* Number on non-redundant bilinearization terms */
  K = mxGetPr(prhs[6]);		/* Matrix of indices k1 and k2 of the rows of F that are combined to produce e new row*/
  D = mxGetPr(prhs[7]);		/* Matrix of indices resulting from the combination of indices k1 and k2 for F*/
  Dg = mxGetPr(prhs[8]);	/* Same, G */
  NDp = mxGetPr(prhs[9]);	/* Size of Dg */
  

  m = *mp;
  Np = *Npp;
  Np1 = Np + 1;
  ND = *NDp;

  num_adds = 0;			
  nptr = 0;                 

  /* Create output vectors*/
  plhs[0] = mxCreateDoubleMatrix(NMAX,Np, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(NMAX,Np, mxREAL);
  /* Indices of non-null bilinearization terms */
  Fi = mxGetPr(plhs[0]);	
  /* Non-null bilinearization terms */
  Fc = mxGetPr(plhs[1]);

  plhs[2] = mxCreateDoubleMatrix(NMAX,Np, mxREAL);
  plhs[3] = mxCreateDoubleMatrix(NMAX,Np, mxREAL);
  plhs[4] = mxCreateDoubleMatrix(1,1, mxREAL);
  NAp = mxGetPr(plhs[4]);

  Gi = mxGetPr(plhs[2]);	
  Gc = mxGetPr(plhs[3]);	



  /* Copy of the first m rows of F */
  mmax =  mxGetM(prhs[0]);	
  for (n=0;n<m;n++){
    for(i=0;i<mmax;i++){
      Fi[nt1=n*NMAX+i]=fi[nt2=n*mmax+i];
      Fc[nt1]=fc[nt2];

    }
  }

  /* Copy of the first m rows of G */
  mmax =  mxGetM(prhs[2]);	
  for (n=0;n<m;n++){
    for(i=0;i<mmax;i++){
      Gi[nt1=n*NMAX+i]=gi[nt2=n*mmax+i];
      Gc[nt1]=gc[nt2];

    }
  }



  /* Loop of columns of Fi (rows of F) */
  n0 = m*NMAX;			/* Initial index in the vectorialization of Fi of the
   * column to compute */

  for (n=m;n<Np;n++){ 		

    root= NULL;			/* Resets binary tree pointer */
    nptr = 0;			/* Resets tree node counter   */

    k1 = K[n+n];		/* Columns to be combined to produce the nth column */
    k2=  K[n+n+1];

    n10 = NMAX*(k1-1); 	/* Initial indices in the vectorialization of Gi of the columns to be combined */
    n20 = NMAX*(k2-1);


    i = n10;
    while ((ft1=Fi[i]) != 0){
      if ((ft = D[((int) ft1)*Np1+k2]) != 0)
	root = addtree(root, ft, Fc[i], &Fi[n0], &Fc[n0]);
     
    i++;                /* Next row */
    }
    i = n20;
    while ((ft1=Fi[i]) != 0){
      if ((ft = D[((int) ft1)*Np1+k1]) != 0)
	root = addtree(root,ft, Fc[i], &Fi[n0], &Fc[n0]);
    i++;                /* Next row */
    }


    n0 += NMAX;		     /* Initial index of next column to be computed */

  }


  /* Loop of columns of Gi (rows of G) */
  n0 = m*NMAX;			/* Initial index in the vectorialization of Fi of the column to compute */

  for (n=m;n<Np;n++){ 		

    root= NULL;			/* Resets binary tree pointer */
    nptr = 0;			/* Resets tree node counter   */

    k1 = K[n+n];		
    k2=  K[n+n+1];

    n10 = NMAX*(k1-1); 		
    n20 = NMAX*(k2-1);

    i = n10;
    while ((ft1=Gi[i]) != 0){
      if ((ft = Dg[((int) ft1-1)*ND+k2]) != 0) 
          /* ft-1 because the elements of Gi correspond to 1 + the order of the expansion term */
	root = addtree(root, ft, Gc[i], &Gi[n0], &Gc[n0]); 
     
    i++;			/* Next row */
    }
    i = n20;
    while ((ft1=Gi[i]) != 0){
      if ((ft = Dg[((int) ft1-1)*ND+k1]) != 0)
	root = addtree(root,ft, Gc[i], &Gi[n0], &Gc[n0]);
    i++;			/* Next row */
    }



    n0 += NMAX;		     /* Initial index of next column to compute */

  }

  *NAp = (double) num_adds;

}
